# methods in cngi_io.image.image should be used going forward,
# methods in cngi_io.image.cngi_image_io are deprecated
from .image import load_image, make_empty_sky_image, read_image, write_image

__all__ = ["load_image", "make_empty_sky_image", "read_image", "write_image"]



from ._zarr.xds_to_zarr import __write_zarr
from ._zarr.xds_from_zarr import __read_zarr
import xarray as xr


def __xds_to_zarr(xds: xr.Dataset, zarr_store: str):
    __write_zarr(xds, zarr_store)


def __xds_from_zarr(zarr_store: str, do_dask: bool):
    # if do_dask is False, all coordinate and data value arrays must
    # be numpy arrays. If True, some may (but aren't required to be)
    # dask arrays
    return __read_zarr(zarr_store, do_dask)



import os
import shutil
import requests
import zipfile

from tqdm import tqdm

FILE_ID = {
    "AA2-Mid-sim_00000.ms": {
        "file": "AA2-Mid-sim_00000.ms.zip",
        "id": "2buf75ebhurjlfhe123qs",
        "rlkey": "8wzgiavzxfp4aza4sx84qanid&dl",
    },
    "Antennae_M8.img.zarr": {
        "file": "Antennae_M8.img.zarr.zip",
        "id": "9v0a1rv7nzm3kqte0u7vp",
        "rlkey": "ws6vd0jbo1dg7jvxsit42q5ba&dl",
    },
    "demo_simulated.im": {
        "file": "demo_simulated.im.zip",
        "id": "z87gibxshwg9e2h155ukk",
        "rlkey": "bn7uvs697wtedij63fa2hu7ed&dl",
    },
    "Antennae_North.cal.lsrk.ms": {
        "file": "Antennae_North.cal.lsrk.ms.zip",
        "id": "olx5qv9avdxxiyjlhlwx2",
        "rlkey": "trrqy43rfcqj4blf9robz4p47&dl",
    },
    "Antennae_North.cal.lsrk.vis.zarr": {
        "file": "Antennae_North.cal.lsrk.vis.zarr.zip",
        "id": "9hcunmq3iqtfiww593nrp",
        "rlkey": "7fingboduee7logszh25n95x5&dl",
    },
    "Antennae_North.cal.lsrk.split.ms": {
        "file": "Antennae_North.cal.lsrk.split.ms.zip",
        "id": "j2e5pd4y7ppvw9efxdfdj",
        "rlkey": "hlb85n40vtac3k9nna14giwsf&dl",
    },
    "Antennae_North.cal.lsrk.split.vis.zarr": {
        "file": "Antennae_North.cal.lsrk.split.ms.zip",
        "id": "hctg3tegkl6ttf2kol0wh",
        "rlkey": "7hxaag4vqk0d674v3368dudad&dl",
    },
    "no_mask.im": {
        "file": "no_mask.zip",
        "id": "4azivw1q7vby4ffawy0tt",
        "rlkey": "91g0hxwx6x4892aisbj5u195z&dl",
    },
}


def download(file, folder="."):
    if os.path.exists("/".join((folder, file))):
        print("File exists.")
        return

    if file not in FILE_ID.keys():
        print("Requested file not found")

        return

    fullname = FILE_ID[file]["file"]
    id = FILE_ID[file]["id"]
    rlkey = FILE_ID[file]["rlkey"]

    url = "https://www.dropbox.com/scl/fi/{id}/{file}?rlkey={rlkey}".format(
        id=id, file=fullname, rlkey=rlkey
    )
    print("url", url)
    headers = {"user-agent": "Wget/1.16 (linux-gnu)"}

    r = requests.get(url, stream=True, headers=headers)
    total = int(r.headers.get("content-length", 0))

    fullname = "/".join((folder, fullname))

    with open(fullname, "wb") as fd, tqdm(
        desc=fullname, total=total, unit="iB", unit_scale=True, unit_divisor=1024
    ) as bar:
        for chunk in r.iter_content(chunk_size=1024):
            if chunk:
                size = fd.write(chunk)
                bar.update(size)

    if zipfile.is_zipfile(fullname):
        shutil.unpack_archive(filename=fullname, extract_dir=folder)

        # Let's clean up after ourselves
        os.remove(fullname)



import logging, os
from typing import Tuple

import numpy as np
import xarray as xr

from casacore import tables


def revert_time(datetimes: np.ndarray) -> np.ndarray:
    """Convert time back from pandas datetime ref to casacore ref
    (reverse of read.convert_casacore_time).

    :param rawtimes: times in pandas reference
    :return: times converted to casacore reference

    """
    return (datetimes.astype(float) / 10**9) + 3506716800.0


#####################################
# translate numpy dtypes to casacore type strings
def type_converter(npdtype: str) -> str:
    cctype = "bad"
    if (npdtype == "int64") or (npdtype == "int32"):
        cctype = "int"
    elif npdtype == "bool":
        cctype = "bool"
    elif npdtype == "float32":
        cctype = "float"
    elif (npdtype == "float64") or (npdtype == "datetime64[ns]"):
        cctype = "double"
    elif npdtype == "complex64":
        cctype = "complex"
    elif npdtype == "complex128":
        cctype = "dcomplex"
    elif str(npdtype).startswith("<U"):
        cctype = "string"

    return cctype


####################################
# create and initialize new output table
def create_table(
    outfile: str,
    xds: xr.Dataset,
    max_rows: int,
    infile=None,
    cols=None,
    generic=False,
):
    if os.path.isdir(outfile):
        os.system("rm -fr %s" % outfile)

    # create column descriptions for table description
    if cols is None:
        cols = list(
            set(list(xds.data_vars) + list(xds.attrs["column_descriptions"].keys()))
            if "column_descriptions" in xds.attrs
            else list(xds.data_vars)
        )
    tabledesc = {}
    for col in cols:
        if ("column_descriptions" in xds.attrs) and (
            col in xds.attrs["column_descriptions"]
        ):
            coldesc = xds.attrs["column_descriptions"][col]
        else:
            coldesc = {"valueType": type_converter(xds[col].dtype)}
            if generic or (
                col == "UVW"
            ):  # will be statically shaped even if not originally
                coldesc = {"shape": tuple(np.clip(xds[col].shape[1:], 1, None))}
            elif xds[col].ndim > 1:  # make variably shaped
                coldesc = {"ndim": xds[col].ndim - 1}
            coldesc["name"] = col
            coldesc["desc"] = col
        tabledesc[col] = coldesc

        # fix the fun set of edge cases from casatestdata that cause errors
        if (tabledesc[col]["dataManagerType"] == "TiledShapeStMan") and (
            tabledesc[col]["ndim"] == 1
        ):
            tabledesc[col]["dataManagerType"] = ""

    if generic:
        tb_tool = tables.table(
            outfile,
            tabledesc=tabledesc,
            nrow=max_rows,
            readonly=False,
            lockoptions={"option": "permanentwait"},
            ack=False,
        )
    else:
        tb_tool = tables.default_ms(outfile, tabledesc)
        tb_tool.addrows(max_rows)
        # if 'DATA_DESC_ID' in cols: tb_tool.putcol('DATA_DESC_ID',
        #    np.zeros((max_rows), dtype='int32') - 1, 0, max_rows)

    # write xds attributes to table keywords, skipping certain reserved attributes
    existing_keywords = tb_tool.getkeywords()
    for attr in xds.attrs:
        if attr in [
            "other",
            "history",
            "info",
        ] + list(existing_keywords.keys()):
            continue
        tb_tool.putkeyword(attr, xds.attrs[attr])
    if "info" in xds.attrs:
        tb_tool.putinfo(xds.attrs["info"])

    # copy subtables and add to main table
    if infile:
        subtables = [
            ss.path
            for ss in os.scandir(infile)
            if ss.is_dir() and ("SORTED_TABLE" not in ss.path)
        ]
        os.system("cp -r %s %s" % (" ".join(subtables), outfile))
        for subtable in subtables:
            if not tables.tableexists(
                os.path.join(outfile, subtable[subtable.rindex("/") + 1 :])
            ):
                continue
            sub_tbl = tables.table(
                os.path.join(outfile, subtable[subtable.rindex("/") + 1 :]),
                readonly=False,
                lockoptions={"option": "permanentwait"},
                ack=False,
            )
            tb_tool.putkeyword(
                subtable[subtable.rindex("/") + 1 :], sub_tbl, makesubrecord=True
            )
            sub_tbl.close()

    tb_tool.close()


############################################################################
##
## write_generic_table() - write any xds to generic casacore table format
##
############################################################################
def write_generic_table(xds: xr.Dataset, outfile: str, subtable="", cols=None):
    """
    Write generic xds contents back to casacore table format on disk

    Parameters
    ----------
    xds : xr.Dataset
        Source xarray dataset data
    outfile : str
        Destination filename (or parent main table if writing subtable)
    subtable : str
        Name of the subtable being written, triggers special logic to add subtable to
        parent table.  Default '' for normal generic writes
    cols : str or list
        List of cols to write. Default None writes all columns
    """
    outfile = os.path.expanduser(outfile)
    logging.debug("writing {os.path.join(outfile, subtable)}")
    if cols is None:
        cols = list(
            set(
                list(xds.data_vars)
                + [cc for cc in xds.coords if cc not in xds.dims]
                + (
                    list(
                        xds.attrs["column_descriptions"].keys()
                        if "column_descriptions" in xds.attrs
                        else []
                    )
                )
            )
        )
    cols = list(np.atleast_1d(cols))

    max_rows = xds.row.shape[0] if "row" in xds.dims else 0
    create_table(
        os.path.join(outfile, subtable),
        xds,
        max_rows,
        infile=None,
        cols=cols,
        generic=True,
    )

    tb_tool = tables.table(
        os.path.join(outfile, subtable),
        readonly=False,
        lockoptions={"option": "permanentwait"},
        ack=False,
    )
    try:
        for dv in cols:
            if (dv not in xds) or (np.prod(xds[dv].shape) == 0):
                continue
            values = (
                xds[dv].values
                if xds[dv].dtype != "datetime64[ns]"
                else revert_time(xds[dv].values)
            )
            tb_tool.putcol(dv, values, 0, values.shape[0], 1)
    except Exception:
        print(
            "ERROR: exception in write generic table - %s, %s, %s, %s"
            % (os.path.join(outfile, subtable), dv, str(values.shape), tb_tool.nrows())
        )

    # now we have to add this subtable to the main table keywords (assuming a main table already exists)
    if len(subtable) > 0:
        main_tbl = tables.table(
            outfile, readonly=False, lockoptions={"option": "permanentwait"}, ack=False
        )
        main_tbl.putkeyword(subtable, tb_tool, makesubrecord=True)
        main_tbl.done()
    tb_tool.close()


###################################
# local helper
def write_main_table_slice(
    xda: xr.DataArray,
    outfile: str,
    ddi: int,
    col: str,
    full_shape: Tuple,
    starts: Tuple,
):
    """
    Write an xds row chunk to the corresponding main table slice
    """
    # trigger the DAG for this chunk and return values while the table is unlocked
    values = xda.compute().values
    if xda.dtype == "datetime64[ns]":
        values = revert_time(values)

    tbs = tables.table(
        outfile, readonly=False, lockoptions={"option": "permanentwait"}, ack=True
    )

    # try:
    if (
        (values.ndim == 1) or (col == "UVW") or (values.shape[1:] == full_shape)
    ):  # scalar columns
        tbs.putcol(col, values, starts[0], len(values))
    else:
        if not tbs.iscelldefined(col, starts[0]):
            tbs.putcell(col, starts[0] + np.arange(len(values)), np.zeros((full_shape)))
        tbs.putcolslice(
            col,
            values,
            starts[1 : values.ndim],
            tuple(np.array(starts[1 : values.ndim]) + np.array(values.shape[1:]) - 1),
            [],
            starts[0],
            len(values),
            1,
        )
    # except:
    #    print("ERROR: write exception - %s, %s, %s" % (col, str(values.shape), str(starts)))
    tbs.close()



#################################
# Helper File
#
# Not exposed in API
#
#################################
import warnings, time, os, logging
import numpy as np
import astropy.wcs
from .common import __get_xds_dim_order, __dask_arrayize
from ._casacore.common import __active_mask
from ._casacore.xds_to_casacore import (
    __coord_dict_from_xds,
    __history_from_xds,
    __imageinfo_dict_from_xds,
    __write_casa_data,
)
from ._casacore.xds_from_casacore import (
    __add_coord_attrs,
    __add_mask,
    __add_sky_or_apeture,
    __casa_image_to_xds_attrs,
    __casa_image_to_xds_metadata,
    __get_mask_names,
    __get_persistent_block,
    __get_starts_shapes_slices,
    __get_transpose_list,
    __make_coord_subset,
    __multibeam_array,
    __read_image_array,
)
from astropy.coordinates import SkyCoord  # High-level coordinates
from astropy.coordinates import ICRS, Galactic, FK4, FK5  # Low-level frames
from astropy.coordinates import Angle, Latitude, Longitude  # Angles
from astropy import units as u
from casacore import tables
from typing import Union
import xarray as xr
from casacore.images import image

warnings.filterwarnings("ignore", category=FutureWarning)


def __load_casa_image_block(infile: str, block_des: dict) -> xr.Dataset:
    image_full_path = os.path.expanduser(infile)
    casa_image = image(image_full_path)
    coords = casa_image.coordinates()
    cshape = casa_image.shape()
    del casa_image
    ret = __casa_image_to_xds_metadata(image_full_path, False)
    xds = ret["xds"]
    starts, shapes, slices = __get_starts_shapes_slices(block_des, coords, cshape)
    xds = __make_coord_subset(xds, slices)
    dimorder = __get_xds_dim_order(ret["sphr_dims"])
    transpose_list, new_axes = __get_transpose_list(coords)
    block = __get_persistent_block(
        image_full_path, shapes, starts, dimorder, transpose_list, new_axes
    )
    xds = __add_sky_or_apeture(xds, block, dimorder, image_full_path, ret["sphr_dims"])
    mymasks = __get_mask_names(image_full_path)
    for m in mymasks:
        full_path = os.sep.join([image_full_path, m])
        block = __get_persistent_block(
            full_path, shapes, starts, dimorder, transpose_list, new_axes
        )
        xds = __add_mask(xds, m, block, dimorder)
    xds.attrs = __casa_image_to_xds_attrs(image_full_path, True)
    mb = __multibeam_array(xds, image_full_path, False)
    if mb is not None:
        selectors = {}
        for k in ("time", "polarization", "frequency"):
            if k in block_des:
                selectors[k] = block_des[k]
        xds["beam"] = mb.isel(selectors)
    xds = __add_coord_attrs(xds, ret["icoords"], ret["dir_axes"])
    return xds


def __read_casa_image(
    infile: str,
    chunks: Union[list, dict],
    masks: bool = True,
    history: bool = True,
    verbose: bool = False,
) -> xr.Dataset:
    img_full_path = os.path.expanduser(infile)
    ret = __casa_image_to_xds_metadata(img_full_path, verbose)
    xds = ret["xds"]
    dimorder = __get_xds_dim_order(ret["sphr_dims"])
    xds = __add_sky_or_apeture(
        xds,
        __read_image_array(img_full_path, chunks, verbose=verbose),
        dimorder,
        img_full_path,
        ret["sphr_dims"],
    )
    if masks:
        mymasks = __get_mask_names(img_full_path)
        for m in mymasks:
            ary = __read_image_array(img_full_path, chunks, mask=m, verbose=verbose)
            xds = __add_mask(xds, m, ary, dimorder)
    xds.attrs = __casa_image_to_xds_attrs(img_full_path, history)
    mb = __multibeam_array(xds, img_full_path, True)
    if mb is not None:
        xds["beam"] = mb
    xds = __add_coord_attrs(xds, ret["icoords"], ret["dir_axes"])
    xds = __dask_arrayize(xds)
    return xds


def __xds_to_casa_image(xds: xr.Dataset, imagename: str) -> None:
    image_full_path = os.path.expanduser(imagename)
    __write_casa_data(xds, image_full_path)
    # create coordinates
    coord = __coord_dict_from_xds(xds)
    ii = __imageinfo_dict_from_xds(xds)
    units = xds.sky.attrs["unit"] if "unit" in xds.sky.attrs else None
    miscinfo = (
        xds.attrs["user"]
        if "user" in xds.attrs and len(xds.attrs["user"]) > 0
        else None
    )
    tb = tables.table(
        image_full_path,
        readonly=False,
        lockoptions={"option": "permanentwait"},
        ack=False,
    )
    tb.putkeyword("coords", coord)
    tb.putkeyword("imageinfo", ii)
    if units:
        tb.putkeyword("units", units)
    if miscinfo:
        tb.putkeyword("miscinfo", miscinfo)
    tb.done()
    # history
    __history_from_xds(xds, image_full_path)



import dask.array as da
import numpy as np
import os
import xarray as xr
from .common import __np_types, __top_level_sub_xds
from ..common import __dask_arrayize, __numpy_arrayize


def __read_zarr(zarr_store: str, do_dask: bool) -> xr.Dataset:
    tmp_xds = xr.open_zarr(zarr_store)
    xds = __decode(tmp_xds, zarr_store, do_dask)
    if do_dask:
        return __dask_arrayize(xds)
    else:
        return __numpy_arrayize(xds)


def __decode(xds: xr.Dataset, zarr_store: str, do_dask: bool) -> xr.Dataset:
    xds.attrs = __decode_dict(xds.attrs, "")
    sub_xdses = __decode_sub_xdses(zarr_store, do_dask)
    for k, v in sub_xdses.items():
        xds.attrs[k] = v
    return xds


def __decode_dict(my_dict: dict, top_key: str) -> dict:
    for k, v in my_dict.items():
        if isinstance(v, dict):
            if (
                "__type" in v
                and v["__type"] == "numpy.ndarray"
                and "__value" in v
                and "__dtype" in v
            ):
                my_dict[k] = np.array(v["__value"], dtype=__np_types[v["__dtype"]])
            else:
                z = os.sep.join([top_key, k]) if top_key else k
                my_dict[k] = __decode_dict(v, z)
    return my_dict


def __decode_sub_xdses(zarr_store: str, do_dask: bool) -> dict:
    sub_xdses = {}
    for root, dirs, files in os.walk(zarr_store):
        for d in dirs:
            if d.startswith(__top_level_sub_xds):
                xds = __read_zarr(os.sep.join([root, d]), do_dask)
                for k, v in xds.data_vars.items():
                    xds = xds.drop_vars([k]).assign({k: v.compute()})
                k = d[len(__top_level_sub_xds) + 2 :]
                sub_xdses[k] = xds
    return sub_xdses



from astropy.wcs import WCS
import numpy as np
import xarray as xr
from typing import Union
from .common import __c


def __make_empty_sky_image(
    xds: xr.Dataset,
    phase_center: Union[list, np.ndarray],
    image_size: Union[list, np.ndarray],
    cell_size: Union[list, np.ndarray],
    chan_coords: Union[list, np.ndarray],
    pol_coords: Union[list, np.ndarray],
    time_coords: Union[list, np.ndarray],
    direction_reference: str,
    projection: str,
    spectral_reference: str,
) -> xr.Dataset:
    if len(image_size) != 2:
        raise ValueError("image_size must have exactly two elements")
    if len(phase_center) != 2:
        raise ValueError("phase_center must have exactly two elements")
    if len(cell_size) != 2:
        raise ValueError("cell_size must have exactly two elements")
    wcs_dict = {}
    wcs_dict["NAXIS1"] = image_size[0]
    wcs_dict["CTYPE1"] = f"RA---{projection}"
    wcs_dict["CRVAL1"] = phase_center[0]
    wcs_dict["CRPIX1"] = image_size[0] // 2 + 1
    wcs_dict["CDELT1"] = -abs(cell_size[0])
    wcs_dict["CUNIT1"] = "rad"
    wcs_dict["NAXIS2"] = image_size[1]
    wcs_dict["CTYPE2"] = f"DEC--{projection}"
    wcs_dict["CRVAL2"] = phase_center[1]
    wcs_dict["CRPIX2"] = image_size[1] // 2 + 1
    wcs_dict["CDELT2"] = abs(cell_size[1])
    wcs_dict["CUNIT2"] = "rad"
    w = WCS(wcs_dict)
    x, y = np.indices(w.pixel_shape)
    long, lat = w.pixel_to_world_values(x, y)
    # long, lat from above eqn will always be in degrees, so convert to rad
    f = np.pi / 180
    long *= f
    lat *= f
    if not isinstance(chan_coords, list) and not isinstance(chan_coords, np.ndarray):
        chan_coords = [chan_coords]
    chan_coords = np.array(chan_coords, dtype=np.float64)
    restfreq = chan_coords[len(chan_coords) // 2]
    vel = (1 - chan_coords / restfreq) * __c
    if not isinstance(time_coords, list) and not isinstance(time_coords, np.ndarray):
        time_coords = [time_coords]
    time_coords = np.array(time_coords, dtype=np.float64)
    coords = {
        "time": time_coords,
        "polarization": pol_coords,
        "frequency": chan_coords,
        "velocity": (("frequency"), vel),
        "right_ascension": (("l", "m"), long),
        "declination": (("l", "m"), lat),
    }
    xds = xds.assign_coords(coords)
    xds.time.attrs = {"format": "MJD", "refer": "UTC", "unit": "d"}
    xds.frequency.attrs = {
        "conversion": {
            "direction": {
                "m0": {"unit": "rad", "value": 0.0},
                "m1": {"unit": "rad", "value": 1.5707963267948966},
                "refer": "FK5",
                "type": "direction",
            },
            "epoch": {
                "m0": {"unit": "d", "value": 0.0},
                "refer": "LAST",
                "type": "epoch",
            },
            "position": {
                "m0": {"unit": "rad", "value": 0.0},
                "m1": {"unit": "rad", "value": 0.0},
                "m2": {"unit": "m", "value": 0.0},
                "refer": "ITRF",
                "type": "position",
            },
            "system": spectral_reference.upper(),
        },
        "native_type": "FREQ",
        "restfreq": restfreq,
        "restfreqs": [restfreq],
        "system": spectral_reference.upper(),
        "unit": "Hz",
        "wave_unit": "mm",
        "wcs": {
            "crval": chan_coords[len(chan_coords) // 2],
            "cdelt": chan_coords[1] - chan_coords[0]
            if len(chan_coords) > 1
            else 1000.0,
            "pc": 1.0,
        },
    }
    xds.velocity.attrs = {"doppler_type": "RADIO", "unit": "m/s"}
    xds.right_ascension.attrs = {
        "unit": "rad",
        "wcs": {"crval": phase_center[0], "cdelt": -abs(cell_size[0])},
    }
    xds.declination.attrs = {
        "unit": "rad",
        "wcs": {"crval": phase_center[1], "cdelt": abs(cell_size[1])},
    }
    xds.attrs = {
        "direction": {
            "conversion_system": direction_reference,
            "conversion_equinox": "J2000",
            "long_pole": 0.0,
            "lat_pole": 0.0,
            #'pc': np.array([[1.0, 0.0], [0.0, 1.0]]),
            "pc": [[1.0, 0.0], [0.0, 1.0]],
            "projection": projection,
            #'projection_parameters': np.array([0.0, 0.0]),
            "projection_parameters": [0.0, 0.0],
            "system": direction_reference,
            "equinox": "J2000",
        },
        "active_mask": "",
        "beam": None,
        "object_name": "",
        "obsdate": {
            "refer": "UTC",
            "format": "MJD",
            "value": time_coords[0],
            "unit": "d",
        },
        "observer": "Karl Jansky",
        #'pointing_center': {
        #    'value': np.array(phase_center), 'initial': True
        # },
        "pointing_center": {"value": list(phase_center), "initial": True},
        "description": "",
        "telescope": {
            "name": "ALMA",
            "position": {
                "type": "position",
                "refer": "ITRF",
                "m2": {"value": 6379946.01326443, "unit": "m"},
                "m1": {"unit": "rad", "value": -0.3994149869262738},
                "m0": {"unit": "rad", "value": -1.1825465955049892},
            },
        },
        "history": None,
    }
    return xds



import pandas as pd


class processing_set(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def summary(self):
        summary_data = {
            "name": [],
            "ddi": [],
            "intent": [],
            "field_id": [],
            "field_name": [],
            "start_frequency": [],
            "end_frequency": [],
        }
        for key, value in self.items():
            summary_data["name"].append(key)
            summary_data["ddi"].append(value.attrs["ddi"])
            summary_data["intent"].append(value.attrs["intent"])
            summary_data["field_id"].append(value.attrs["field_info"]["field_id"])
            summary_data["field_name"].append(value.attrs["field_info"]["name"])
            summary_data["start_frequency"].append(value["frequency"].values[0])
            summary_data["end_frequency"].append(value["frequency"].values[-1])
        summary_df = pd.DataFrame(summary_data)
        return summary_df

    def get(self, id):
        return self[list(self.keys())[id]]



from .read_processing_set import read_processing_set
from .load_processing_set import load_processing_set
from .convert_msv2_to_processing_set import convert_msv2_to_processing_set

from .vis_io import read_vis, load_vis_block, write_vis

__all__ = ["read_vis", "load_vis_block", "write_vis"]



# from numcodecs.zstd import Zstd
import numcodecs
from typing import Dict, Tuple, Union
import itertools
import json
import numbers

import dask
from xradio.vis._vis_utils._ms.descr import describe_ms
from xradio.vis._vis_utils._ms.msv2_msv3 import ignore_msv2_cols
from xradio.vis._vis_utils._ms._tables.read import (
    read_generic_table,
    convert_casacore_time,
    extract_table_attributes,
)
from xradio.vis._vis_utils._ms._tables.read_main_table import (
    get_baselines,
    get_utimes_tol,
)
from xradio.vis._vis_utils._ms.subtables import (
    subt_rename_ids,
)
from xradio.vis._vis_utils._ms._tables.table_query import open_table_ro, open_query
from xradio.vis._vis_utils._utils.stokes_types import stokes_types
import numpy as np
import logging
import time
import xarray as xr


from ._zarr import add_encoding
from ._msv2_to_msv4_meta import (
    column_description_casacore_to_msv4_measure,
    create_attribute_metadata,
    col_to_data_variable_names,
    col_dims,
)

from .read_processing_set import read_processing_set


def calc_indx_for_row_split(tb_tool, taql_where):
    baselines = get_baselines(tb_tool)
    col_names = tb_tool.colnames()
    cshapes = [
        np.array(tb_tool.getcell(col, 0)).shape
        for col in col_names
        if tb_tool.iscelldefined(col, 0)
    ]

    freq_cnt, pol_cnt = [(cc[0], cc[1]) for cc in cshapes if len(cc) == 2][0]
    utimes, tol = get_utimes_tol(tb_tool, taql_where)
    # utimes = np.unique(tb_tool.getcol("TIME"))

    tvars = {}
    tidxs = np.searchsorted(utimes, tb_tool.getcol("TIME"))

    ts_ant1, ts_ant2 = (
        tb_tool.getcol("ANTENNA1"),
        tb_tool.getcol("ANTENNA2"),
    )

    ts_bases = [
        str(ll[0]).zfill(3) + "_" + str(ll[1]).zfill(3)
        for ll in np.hstack([ts_ant1[:, None], ts_ant2[:, None]])
    ]
    bidxs = np.searchsorted(baselines, ts_bases)

    # some antenna 2"s will be out of bounds for this chunk, store rows that are in bounds
    didxs = np.where((bidxs >= 0) & (bidxs < len(baselines)))[0]

    baseline_ant1_id, baseline_ant2_id = np.array(
        [tuple(map(int, x.split("_"))) for x in baselines]
    ).T
    return (
        tidxs,
        bidxs,
        didxs,
        baseline_ant1_id,
        baseline_ant2_id,
        convert_casacore_time(utimes, False),
    )


def check_if_consistent(col, col_name):
    """_summary_

    Parameters
    ----------
    col : _type_
        _description_
    col_name : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    col_unique = np.unique(col)
    assert len(col_unique) == 1, col_name + " is not consistent."
    return col_unique[0]


def read_col(
    tb_tool,
    col: str,
    cshape: Tuple[int],
    tidxs: np.ndarray,
    bidxs: np.ndarray,
    didxs: np.ndarray,
):
    data = tb_tool.getcol(col)
    fulldata = np.full(cshape + data.shape[1:], np.nan, dtype=data.dtype)
    fulldata[tidxs, bidxs] = data
    return fulldata


def create_coordinates(
    xds, in_file, ddi, utime, interval, baseline_ant1_id, baseline_ant2_id
):
    coords = {
        "time": utime,
        "baseline_antenna1_id": ("baseline_id", baseline_ant1_id),
        "baseline_antenna2_id": ("baseline_id", baseline_ant2_id),
        "uvw_label": ["u", "v", "w"],
        "baseline_id": np.arange(len(baseline_ant1_id)),
    }

    ddi_xds = read_generic_table(in_file, "DATA_DESCRIPTION").sel(row=ddi)
    pol_setup_id = ddi_xds.polarization_id.values
    spw_id = ddi_xds.spectral_window_id.values

    spw_xds = read_generic_table(
        in_file,
        "SPECTRAL_WINDOW",
        rename_ids=subt_rename_ids["SPECTRAL_WINDOW"],
    ).sel(spectral_window_id=spw_id)
    coords["frequency"] = spw_xds["chan_freq"].data[
        ~(np.isnan(spw_xds["chan_freq"].data))
    ]

    pol_xds = read_generic_table(
        in_file,
        "POLARIZATION",
        rename_ids=subt_rename_ids["POLARIZATION"],
    )
    num_corr = int(pol_xds["num_corr"][pol_setup_id].values)
    coords["polarization"] = np.vectorize(stokes_types.get)(
        pol_xds["corr_type"][pol_setup_id, :num_corr].values
    )

    xds = xds.assign_coords(coords)

    ###### Create Frequency Coordinate ######
    freq_column_description = spw_xds.attrs["other"]["msv2"]["ctds_attrs"][
        "column_descriptions"
    ]

    msv4_measure = column_description_casacore_to_msv4_measure(
        freq_column_description["CHAN_FREQ"], ref_code=spw_xds["meas_freq_ref"].data
    )
    xds.frequency.attrs.update(msv4_measure)

    xds.frequency.attrs["spectral_window_name"] = str(spw_xds.name.values)
    msv4_measure = column_description_casacore_to_msv4_measure(
        freq_column_description["REF_FREQUENCY"], ref_code=spw_xds["meas_freq_ref"].data
    )
    xds.frequency.attrs["reference_frequency"] = {
        "dims": "",
        "data": float(spw_xds.ref_frequency.values),
        "attrs": msv4_measure,
    }
    xds.frequency.attrs["spw_id"] = spw_id

    # xds.frequency.attrs["effective_channel_width"] = "EFFECTIVE_CHANNEL_WIDTH"
    # Add if doppler table is present
    # xds.frequency.attrs["doppler_velocity"] =
    # xds.frequency.attrs["doppler_type"] =
    unique_chan_width = np.unique(
        spw_xds.chan_width.data[np.logical_not(np.isnan(spw_xds.chan_width.data))]
    )
    # assert len(unique_chan_width) == 1, "Channel width varies for spw."
    # xds.frequency.attrs["channel_width"] = spw_xds.chan_width.data[
    #    ~(np.isnan(spw_xds.chan_width.data))
    # ]  # unique_chan_width[0]
    msv4_measure = column_description_casacore_to_msv4_measure(
        freq_column_description["CHAN_WIDTH"], ref_code=spw_xds["meas_freq_ref"].data
    )
    if not msv4_measure:
        msv4_measure["type"] = "quantity"
        msv4_measure["units"] = ["Hz"]
    xds.frequency.attrs["channel_width"] = {
        "dims": "",
        "data": np.abs(unique_chan_width[0]),
        "attrs": msv4_measure,
    }

    ###### Create Time Coordinate ######
    main_table_attrs = extract_table_attributes(in_file)
    main_column_descriptions = main_table_attrs["column_descriptions"]
    msv4_measure = column_description_casacore_to_msv4_measure(
        main_column_descriptions["TIME"]
    )
    xds.time.attrs.update(msv4_measure)

    msv4_measure = column_description_casacore_to_msv4_measure(
        main_column_descriptions["INTERVAL"]
    )
    if not msv4_measure:
        msv4_measure["type"] = "quantity"
        msv4_measure["units"] = ["s"]
    xds.time.attrs["integration_time"] = {
        "dims": "",
        "data": interval,
        "attrs": msv4_measure,
    }
    xds.time.attrs["effective_integration_time"] = "EFFECTIVE_INTEGRATION_TIME"
    return xds


def create_taql_query_and_file_name(out_file, intent, state_ids, field_id, ddi):
    file_name = (
        out_file
        + "/"
        + out_file.replace(".vis.zarr", "").split("/")[-1]
        + "_ddi_"
        + str(ddi)
        + "_intent_"
        + intent
    )

    taql_where = f"where (DATA_DESC_ID = {ddi})"

    if isinstance(state_ids, numbers.Integral):
        taql_where += f" AND (STATE_ID = {state_ids})"
    else:
        state_ids_or = " OR STATE_ID = ".join(np.char.mod("%d", state_ids))
        taql_where += f" AND (STATE_ID = {state_ids_or})"

    if field_id is not None:
        taql_where += f" AND (FIELD_ID = {field_id})"
        file_name = file_name + "_field_id_" + str(field_id)

    return taql_where, file_name


def create_data_variables(
    in_file, xds, tb_tool, time_baseline_shape, tidxs, bidxs, didxs
):
    # Create Data Variables
    col_names = tb_tool.colnames()

    main_table_attrs = extract_table_attributes(in_file)
    main_column_descriptions = main_table_attrs["column_descriptions"]
    for col in col_names:
        if col in col_to_data_variable_names:
            if (col == "WEIGHT") and ("WEIGHT_SPECTRUM" in col_names):
                continue
            try:
                start = time.time()
                if col == "WEIGHT":
                    xds[col_to_data_variable_names[col]] = xr.DataArray(
                        np.tile(
                            read_col(
                                tb_tool,
                                col,
                                time_baseline_shape,
                                tidxs,
                                bidxs,
                                didxs,
                            )[:, :, None, :],
                            (1, 1, xds.dims["frequency"], 1),
                        ),
                        dims=col_dims[col],
                    )

                else:
                    xds[col_to_data_variable_names[col]] = xr.DataArray(
                        read_col(
                            tb_tool,
                            col,
                            time_baseline_shape,
                            tidxs,
                            bidxs,
                            didxs,
                        ),
                        dims=col_dims[col],
                    )
                    # logging.info("Time to read column " + str(col) + " : " + str(time.time()-start))
            except:
                # logging.debug("Could not load column",col)
                continue

            xds[col_to_data_variable_names[col]].attrs.update(
                create_attribute_metadata(col, main_column_descriptions)
            )


def convert_and_write_partition(
    in_file: str,
    out_file: str,
    intent: str,
    ddi: int = 0,
    state_ids=None,
    field_id: int = None,
    ignore_msv2_cols: Union[list, None] = None,
    main_chunksize: Union[Dict, None] = None,
    compressor: numcodecs.abc.Codec = numcodecs.Zstd(level=2),
    storage_backend="zarr",
    overwrite: bool = False,
):
    """_summary_

    Parameters
    ----------
    in_file : str
        _description_
    out_file : str
        _description_
    intent : str
        _description_
    ddi : int, optional
        _description_, by default 0
    state_ids : _type_, optional
        _description_, by default None
    field_id : int, optional
        _description_, by default None
    ignore_msv2_cols : Union[list, None], optional
        _description_, by default None
    main_chunksize : Union[Dict, None], optional
        _description_, by default None
    compressor : numcodecs.abc.Codec, optional
        _description_, by default numcodecs.Zstd(level=2)
    storage_backend : str, optional
        _description_, by default "zarr"
    overwrite : bool, optional
        _description_, by default False

    Returns
    -------
    _type_
        _description_
    """
    if ignore_msv2_cols is None:
        ignore_msv2_cols = []

    taql_where, file_name = create_taql_query_and_file_name(
        out_file, intent, state_ids, field_id, ddi
    )

    start_with = time.time()
    with open_table_ro(in_file) as mtable:
        taql_main = f"select * from $mtable {taql_where}"
        with open_query(mtable, taql_main) as tb_tool:
            if tb_tool.nrows() == 0:
                tb_tool.close()
                mtable.close()
                return xr.Dataset(), {}, {}

            (
                tidxs,
                bidxs,
                didxs,
                baseline_ant1_id,
                baseline_ant2_id,
                utime,
            ) = calc_indx_for_row_split(tb_tool, taql_where)
            time_baseline_shape = (len(utime), len(baseline_ant1_id))
            # logging.debug("Calc indx for row split "+ str(time.time()-start))

            xds = xr.Dataset()
            interval = check_if_consistent(tb_tool.getcol("INTERVAL"), "INTERVAL")
            xds = create_coordinates(
                xds, in_file, ddi, utime, interval, baseline_ant1_id, baseline_ant2_id
            )

            create_data_variables(
                in_file, xds, tb_tool, time_baseline_shape, tidxs, bidxs, didxs
            )

            # Create field_info
            field_id = check_if_consistent(tb_tool.getcol("FIELD_ID"), "FIELD_ID")
            xds = create_field_info(xds, in_file, field_id)

            # Create ant_xds
            ant_xds = create_ant_xds(in_file)

            # To do: add other _info and _xds creation functions.

            # Fix UVW frame
            # From CASA fixvis docs: clean and the im tool ignore the reference frame claimed by the UVW column (it is often mislabelled as ITRF when it is really FK5 (J2000)) and instead assume the (u, v, w)s are in the same frame as the phase tracking center. calcuvw does not yet force the UVW column and field centers to use the same reference frame! Blank = use the phase tracking frame of vis.
            xds.UVW.attrs["frame"] = xds.attrs["field_info"]["phase_direction"][
                "attrs"
            ]["frame"]

            xds.attrs["intent"] = intent
            xds.attrs["ddi"] = ddi

            # Time and frequency should always be increasing
            if len(xds.frequency) > 1 and xds.frequency[1] - xds.frequency[0] < 0:
                xds = xds.sel(frequency=slice(None, None, -1))

            if len(xds.time) > 1 and xds.time[1] - xds.time[0] < 0:
                xds = xds.sel(time=slice(None, None, -1))

            # Add data_groups
            xds.attrs["data_groups"] = {
                "base": {
                    "visibility": "VISIBILITY",
                    "flag": "FLAG",
                    "weight": "WEIGHT",
                    "uvw": "UVW",
                }
            }

            if overwrite:
                mode = "w"
            else:
                mode = "w-"

            add_encoding(xds, compressor=compressor, chunks=main_chunksize)

            if storage_backend == "zarr":
                xds.to_zarr(store=file_name + "/MAIN", mode=mode)
                ant_xds.to_zarr(store=file_name + "/ANTENNA", mode=mode)
            elif storage_backend == "netcdf":
                # xds.to_netcdf(path=file_name+"/MAIN", mode=mode) #Does not work
                raise

    # logging.info("Saved ms_v4 " + file_name + " in " + str(time.time() - start_with) + "s")


def create_ant_xds(in_file):
    to_new_data_variable_names = {
        "position": "POSITION",
        "offset": "FEED_OFFSET",
        "dish_diameter": "DISH_DIAMETER",
    }
    ant_dims = {
        "position": ["antenna_id", "coordinate"],
        "offset": ["antenna_id", "coordinate"],
        "dish_diameter": ["antenna_id"],
    }
    to_new_coord_names = {
        "name": "name",
        "station": "station",
        "type": "type",
        "mount": "mount",
        "phased_array_id": "phased_array_id",
    }

    generic_ant_xds = read_generic_table(
        in_file,
        "ANTENNA",
        rename_ids=subt_rename_ids["ANTENNA"],
    )

    ant_column_description = generic_ant_xds.attrs["other"]["msv2"]["ctds_attrs"][
        "column_descriptions"
    ]
    # ['OFFSET', 'POSITION', 'TYPE', 'DISH_DIAMETER', 'FLAG_ROW', 'MOUNT', 'NAME', 'STATION']
    ant_xds = xr.Dataset()

    coords = {"antenna_id": np.arange(generic_ant_xds.dims["antenna_id"])}
    for key in generic_ant_xds:
        msv4_measure = column_description_casacore_to_msv4_measure(
            ant_column_description[key.upper()]
        )
        if key in to_new_data_variable_names:
            ant_xds[to_new_data_variable_names[key]] = xr.DataArray(
                generic_ant_xds[key].data, dims=ant_dims[key]
            )

            if msv4_measure:
                ant_xds[to_new_data_variable_names[key]].attrs.update(msv4_measure)

            if key in ["dish_diameter"]:
                ant_xds[to_new_data_variable_names[key]].attrs.update(
                    {"units": ["m"], "type": "quantity"}
                )

        if key in to_new_coord_names:
            coords[to_new_coord_names[key]] = ("antenna_id", generic_ant_xds[key].data)

    ant_xds = ant_xds.assign_coords(coords)
    return ant_xds


def create_field_info(xds, in_file, field_id):
    field_xds = read_generic_table(
        in_file,
        "FIELD",
        rename_ids=subt_rename_ids["FIELD"],
    ).sel(field_id=field_id)
    # https://stackoverflow.com/questions/53195684/how-to-navigate-a-dict-by-list-of-keys

    field_column_description = field_xds.attrs["other"]["msv2"]["ctds_attrs"][
        "column_descriptions"
    ]
    # ['DELAY_DIR', 'PHASE_DIR', 'REFERENCE_DIR', 'CODE', 'FLAG_ROW', 'NAME', 'NUM_POLY', 'SOURCE_ID', 'TIME']

    msv4_measure = column_description_casacore_to_msv4_measure(
        field_column_description["REFERENCE_DIR"],
        ref_code=getattr(field_xds.get("refdir_ref"), "data", None),
    )
    delay_dir = {
        "dims": "",
        "data": list(field_xds["delay_dir"].data[0, :]),
        "attrs": msv4_measure,
    }

    msv4_measure = column_description_casacore_to_msv4_measure(
        field_column_description["PHASE_DIR"],
        ref_code=getattr(field_xds.get("phasedir_ref"), "data", None),
    )
    phase_dir = {
        "dims": "",
        "data": list(field_xds["phase_dir"].data[0, :]),
        "attrs": msv4_measure,
    }

    msv4_measure = column_description_casacore_to_msv4_measure(
        field_column_description["DELAY_DIR"],
        ref_code=getattr(field_xds.get("delaydir_ref"), "data", None),
    )
    reference_dir = {
        "dims": "",
        "data": list(field_xds["delay_dir"].data[0, :]),
        "attrs": msv4_measure,
    }

    field_info = {
        "name": str(field_xds["name"].data),
        "code": str(field_xds["code"].data),
        "delay_direction": delay_dir,
        "phase_direction": phase_dir,
        "reference_direction": reference_dir,
        "field_id": field_id,
    }
    xds.attrs["field_info"] = field_info

    return xds


def get_unqiue_intents(in_file):
    """_summary_

    Parameters
    ----------
    in_file : str
        _description_

    Returns
    -------
    _type_
        _description_
    """
    state_xds = read_generic_table(
        in_file,
        "STATE",
        rename_ids=subt_rename_ids["STATE"],
    )

    if len(state_xds.data_vars) > 0:
        obs_mode_dict = {}
        for i, obs_mode in enumerate(state_xds.obs_mode.values):
            if obs_mode in obs_mode_dict:
                obs_mode_dict[obs_mode].append(i)
            else:
                obs_mode_dict[obs_mode] = [i]
        return list(obs_mode_dict.keys()), list(obs_mode_dict.values())
    else:  # empty state table
        return ["None"], [0]


def enumerated_product(*args):
    yield from zip(
        itertools.product(*(range(len(x)) for x in args)), itertools.product(*args)
    )


def create_partition_enumerated_product(in_file: str, partition_scheme: str):
    """Creates an enumerated_product of the data_desc_ids, state_ids, field_ids in a MS v2 that define the partions in a processing set.

    Parameters
    ----------
    in_file : str
        _description_
    partition_scheme : str
        _description_

    Returns
    -------
    _type_
        _description_
    """
    spw_xds = read_generic_table(
        in_file,
        "SPECTRAL_WINDOW",
        rename_ids=subt_rename_ids["SPECTRAL_WINDOW"],
    )

    ddi_xds = read_generic_table(in_file, "DATA_DESCRIPTION")
    data_desc_ids = np.arange(ddi_xds.dims["row"])

    if partition_scheme == "ddi_intent_field":
        intents, state_ids = get_unqiue_intents(in_file)
        field_ids = np.arange(read_generic_table(in_file, "FIELD").dims["row"])
    elif partition_scheme == "ddi_state_field":
        state_xds = read_generic_table(in_file, "STATE")

        if len(state_xds.data_vars) > 0:
            state_ids = [np.arange(state_xds.dims["row"])]
            intents = state_xds.obs_mode.values
        else:  # empty state table
            state_ids = [0]
            intents = ["None"]
        # print(state_xds, intents)
        # field_ids = [None]
        field_ids = np.arange(read_generic_table(in_file, "FIELD").dims["row"])

    return enumerated_product(data_desc_ids, state_ids, field_ids), intents


def convert_msv2_to_processing_set(
    in_file: str,
    out_file: str,
    partition_scheme: str = "intent_field",
    main_chunksize: Union[Dict, str, None] = None,
    pointing_chunksize: Union[Dict, str, None] = None,
    compressor: numcodecs.abc.Codec = numcodecs.Zstd(level=2),
    storage_backend="zarr",
    parallel: bool = False,
    overwrite: bool = False,
):
    """Convert a Measurement Set v2 into a Processing Set of Measurement Set v4.

    Parameters
    ----------
    in_file : str
        Input MS name.
    out_file : str
        Output PS name.
    partition_scheme : {"ddi_intent_field", "ddi_state_field"}, optional
        A MS v4 can only contain a single spectral window, polarization setup, intent, and field. Consequently, the MS v2 is partitioned when converting to MS v4.
        The partition_scheme "ddi_intent_field" gives the largest partition that meets the MS v4 specification. The partition_scheme "ddi_state_field" gives a finer granularity where the data is also partitioned by state (the state partitioning will ensure a single intent).
        By default, "ddi_intent_field".
    main_chunksize : Union[Dict, str, None], optional
        A dictionary that defines the chunk size of the main dataset. Acceptable keys are "time", "baseline", "antenna", "frequency", "polarization". By default, None.
    pointing_chunksize : Union[Dict, str, None], optional
        A dictionary that defines the chunk size of the pointing dataset. Acceptable keys are "time", "antenna", "polarization". By default, None.
    compressor : numcodecs.abc.Codec, optional
        The Blosc compressor to use when saving the converted data to disk using Zarr, by default numcodecs.Zstd(level=2).
    storage_backend : {"zarr", "netcdf"}, optional
        The on-disk format to use. "netcdf" is not yet implemented.
    parallel : bool, optional
        Makes use of Dask to execute conversion in parallel, by default False.
    overwrite : bool, optional
        Whether to overwrite an existing processing set, by default False.
    """

    partition_enumerated_product, intents = create_partition_enumerated_product(
        in_file, partition_scheme
    )

    delayed_list = []
    for idx, pair in partition_enumerated_product:
        ddi, state_id, field_id = pair
        logging.debug(
            "DDI " + str(ddi) + ", STATE " + str(state_id) + ", FIELD " + str(field_id)
        )

        if partition_scheme == "ddi_intent_field":
            intent = intents[idx[1]]
        else:
            intent = intents[idx[1]] + "_" + str(state_id)

        if parallel:
            delayed_list.append(
                dask.delayed(convert_and_write_partition)(
                    in_file,
                    out_file,
                    intent,
                    ddi,
                    state_id,
                    field_id,
                    ignore_msv2_cols=ignore_msv2_cols,
                    main_chunksize=main_chunksize,
                    compressor=compressor,
                    overwrite=overwrite,
                )
            )
        else:
            convert_and_write_partition(
                in_file,
                out_file,
                intent,
                ddi,
                state_id,
                field_id,
                ignore_msv2_cols=ignore_msv2_cols,
                main_chunksize=main_chunksize,
                compressor=compressor,
                storage_backend=storage_backend,
                overwrite=overwrite,
            )

    if parallel:
        dask.compute(delayed_list)



import astropy.units as u
import dask
import dask.array as da
import numpy as np
import xarray as xr


__c = 2.99792458e08 * u.m / u.s
# OPTICAL = Z
__doppler_types = ["RADIO", "Z", "RATIO", "BETA", "GAMMA"]
__image_type = "image_type"


def __get_xds_dim_order(has_sph: bool) -> list:
    dimorder = ["time", "polarization", "frequency"]
    dir_lin = ["l", "m"] if has_sph else ["u", "v"]
    dimorder.extend(dir_lin)
    return dimorder


def __coords_to_numpy(xds):
    for k, v in xds.coords.items():
        if dask.is_dask_collection(v):
            attrs = xds[k].attrs
            xds = xds.assign_coords({k: (v.dims, v.to_numpy())})
            xds[k].attrs = attrs
    return xds


def __dask_arrayize(xds):
    """
    If necessary, change coordinates to numpy arrays and data
    variables to dask arrays
    """
    xds = __coords_to_numpy(xds)
    for k, v in xds.data_vars.items():
        if not dask.is_dask_collection(v):
            attrs = xds[k].attrs
            xds = xds.drop_vars([k])
            # may need to add dims to this call as in numpy method analogs in this file
            xds = xds.assign({k: da.array(v)})
            xds[k].attrs = attrs
    for k, v in xds.attrs.items():
        if isinstance(v, xr.Dataset):
            xds.attrs[k] = __dask_arrayize(v)
    return xds


def __numpy_arrayize(xds):
    xds = __coords_to_numpy(xds)
    for k, v in xds.data_vars.items():
        if dask.is_dask_collection(v):
            attrs = xds[k].attrs
            xds = xds.drop_vars([k])
            xds = xds.assign({k: (v.dims, v.to_numpy())})
            xds[k].attrs = attrs
    for k, v in xds.attrs.items():
        if isinstance(v, xr.Dataset):
            xds.attrs[k] = __dask_arrayize(v)
    return xds


def __default_freq_info() -> dict():
    return {
        "conversion": {
            "direction": {
                "m0": {"unit": "rad", "value": 0.0},
                "m1": {"unit": "rad", "value": 1.5707963267948966},
                "system": "FK5",
                "equinox": "J2000",
            },
            "epoch": {
                "m0": {"unit": "d", "value": 0.0},
                "refer": "LAST",
            },
            "position": {
                "m0": {"unit": "rad", "value": 0.0},
                "m1": {"unit": "rad", "value": 0.0},
                "m2": {"unit": "m", "value": 0.0},
                "refer": "ITRF",
            },
            "system": "LSRK",
        },
        "nativeType": "FREQ",
        "restfreq": 1420405751.7860003,
        "restfreqs": [1420405751.7860003],
        "system": "LSRK",
        "unit": "Hz",
        "waveUnit": "mm",
        "wcs": {
            "cdelt": 1000.0,
            "crval": 1415000000.0,
        },
    }


def __freq_from_vel(
    crval: float,
    cdelt: float,
    crpix: float,
    cunit: str,
    ctype: str,
    nchan: float,
    restfreq: float,
) -> tuple:
    """
    inputs are velocity info
    restfreq is an astropy quantity with frequency units
    a tuple of two dicts is returned. The first dict represents
    the frequencies, the second the velocities. Both dicts have
    keys 'value', 'unit', 'crval', 'cdelt', 'crpix'. 'value'
    is a list.
    """
    v0 = crval - cdelt * crpix
    vel = [v0 + i * cdelt for i in range(nchan)]
    vel = vel * u.Unit(cunit)
    v_dict = {
        "value": vel.value,
        "unit": cunit,
        "crval": crval,
        "cdelt": cdelt,
        "crpix": crpix,
    }
    uctype = ctype.upper()
    if uctype == "Z" or uctype == "OPTICAL":
        freq = restfreq / (np.array(vel.value) * vel.unit / __c + 1)
        freq = freq.to(u.Hz)
        fcrval = restfreq / (crval * vel.unit / __c + 1)
        fcdelt = -restfreq / __c / (crval * vel.unit / __c + 1) ** 2 * cdelt * vel.unit
        f_dict = {
            "value": freq.value,
            "unit": "Hz",
            "crval": fcrval.to(u.Hz).value,
            "cdelt": fcdelt.to(u.Hz).value,
            "crpix": crpix,
        }
    else:
        raise RuntimeError(f"Unhandled doppler type {ctype}")
    return f_dict, v_dict



import numcodecs, os
from typing import Dict, List, Tuple, Union

import xarray as xr

from ._vis_utils._utils.cds import CASAVisSet
from xradio.vis import _vis_utils


def read_vis(
    infile: str,
    subtables: bool = True,
    asdm_subtables: bool = False,
    partition_scheme: str = "intent",
    chunks: Union[Tuple[int], List[int]] = None,
    expand: bool = False,
) -> CASAVisSet:
    """Read a MeasurementSet (MSv2 format) into a next generation CASA
    dataset (visibilities dataset as a set of Xarray datasets).

    The MS is partitioned into multiple sub- Xarray datasets (where the data variables are read as
    Dask delayed arrays).
    The MS is partitioned by DDI, which guarantees a fixed data shape per partition (in terms of channels
    and polarizations) and, subject to experimentation, by scan and subscan. This results in multiple
    partitions as xarray datasets (xds) contained within a main xds (mxds).

    :param infile: Input MS filename
    :param rowmap: (to be removed) Dictionary of DDI to tuple of (row indices, channel indices). Returned
    by ms_selection function. Default None ignores selections
    :param subtables: Also read and include subtables along with main table selection. Default False will
    omit subtables (faster)
    :param asdm_subtables: in addition to MeasurementSet subtables (if enabled), also read extension
    subtables names "ASDM_*"
    :param partition_scheme: (experimenting) Whether to partition sub-xds datasets by scan/subscan
    (in addition to DDI), or other alternative partitioning schemes. Accepted values: 'scan/subscan',
    'scan', 'ddi', 'intent'. Default: 'intent'
    :param chunks: Can be used to set a specific chunk shape (with a tuple of ints), or to control the
    optimization used for automatic chunking (with a list of ints). A tuple of ints in the form of (row,
    chan, pol) will use a fixed chunk shape. A list or numpy array of ints in the form of [idx1, etc]
    will trigger auto-chunking optimized for the given indices, with row=0, chan=1, pol=2. Default None
    uses auto-chunking with a best fit across all dimensions (probably sub-optimal for most cases).
    :param expand: (to be removed) Whether or not to return the original flat row structure of the MS (False)
    or expand the rows to time x baseline dimensions (True). Expanding the rows allows for easier indexing
    and parallelization across time and baseline dimensions, at the cost of some conversion time. Default
    False
    :param **kwargs: (to be removed?) Selection parameters from the standard way of making CASA MS
    selections. Supported keys are: spw, field, scan, baseline, time, scanintent, uvdist, polarization,
    array, observation.  Values are strings.

    :return: ngCASA visisbilities dataset, essentially made of two dictionaries of
    metainformation and data partitions
    """
    infile = os.path.expanduser(infile)
    if not os.path.isdir(infile):
        raise ValueError(f"invalid input filename to read_vis {infile}")

    if _vis_utils.zarr.is_zarr_vis(infile):
        return _vis_utils.zarr.read_vis(infile, subtables, asdm_subtables)
    else:
        return _vis_utils.ms.read_ms(
            infile, subtables, asdm_subtables, partition_scheme, chunks, expand
        )


def load_vis_block(
    infile: str,
    block_des: Dict[str, slice],
    partition_key: Tuple[int, int, str],
    subtables: List[str] = None,
) -> Dict[Tuple[int, int], xr.Dataset]:
    """Read a chunk of a visibilities dataset into an Xarray dataset, loading the
    data in memory.
    The input format support is the MeasurementSet v2 (MSv2 format)

    :param infile: Input visibilities path
    :param block_des: specification of chunk to load

    :return: CASA visibilities dataset holding a chunk of visibility data, for one
    partition
    (spw_id, pol_setup_id, intent_string triplet)
    """
    # TODO: use the input partition_key
    # the intent str of the partition_key is not yet effectively used
    # TODO: support subtables list
    return _vis_utils.ms.load_vis_chunk(infile, block_des, partition_key)


def write_vis(
    cds: CASAVisSet,
    outpath: str,
    chunks_on_disk: Union[Dict, None] = None,
    compressor: Union[numcodecs.abc.Codec, None] = None,
    out_format: str = "zarr",
) -> None:
    """Write CASA vis dataset to disk.
    The disk format supported is "zarr". When chunks_on_disk is not specified the
    chunking in the input dataset is used. When chunks_on_disk is specified that
    dataset is saved using that chunking.

    :param cds: CASA visibilities dataset to write to disk
    :param outpath: output path, generally ends in .zarr
    :param chunks_on_disk: a dictionary with the chunk size that will
    be used when writing to disk. For example {'time': 20, 'chan': 6}.
    If chunks_on_disk is not specified the chunking of dataset will
    be used.
    :param compressor: the blosc compressor to use when saving the
    converted data to disk using zarr. If None the zstd compression
    algorithm used with compression level 2.
    """

    if out_format == "zarr":
        return _vis_utils.zarr.write_vis(cds, outpath, chunks_on_disk, compressor)
    else:
        raise ValueError(f"Unsupported output format: {out_format}")
